/*
 *
 * (C) COPYRIGHT 2014-2017 MStar Semiconductor, Inc. All rights reserved.
 *
 * This program is free software and is provided to you under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation, and any use by you of this program is subject to the terms
 * of such GNU licence.
 *
 * A copy of the licence is included with the program, and can also be obtained
 * from Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 *
 */

#include <mali_kbase.h>
#include "mali_kbase_platform_mstar.h"
#include "mali_kbase_platform_mstar_m7621_reg.h"

#include <linux/delay.h>
#include "mstar/mstar_chip.h"

/* Platform functions */
void init_registers(void)
{
#define MIU_BASE_TO_MB(base)        ((base) >> 20)                   /* in MB */
#define DRAM_SIZE_REG_TO_MB(reg)    ((reg > 0) ? (1 << (reg)) : 0)   /* in MB */

    /* set GPU clock */
    SET_REG(gpu_pll, set_lo, (MHZ_TO_REG(MALI_MAX_FREQ)>> 0)&0xffff);
    SET_REG(gpu_pll, set_hi, (MHZ_TO_REG(MALI_MAX_FREQ)>>16)&0x00ff);
    SET_REG(gpu_pll, sld, 1);
    udelay(10);

    /* set MIU addr */
    SET_REG(gpu, miu0_start, MIU_BASE_TO_MB(MSTAR_MIU0_BUS_BASE));
    SET_REG(gpu, miu0_end, MIU_BASE_TO_MB(MSTAR_MIU0_BUS_BASE) + DRAM_SIZE_REG_TO_MB(GET_REG(miu0, dram_size)));
#if defined(MSTAR_MIU1_BUS_BASE) && defined(ARM_MIU1_BASE_ADDR)
    SET_REG(gpu, miu1_start, MIU_BASE_TO_MB(MSTAR_MIU1_BUS_BASE));
    SET_REG(gpu, miu1_end, MIU_BASE_TO_MB(MSTAR_MIU1_BUS_BASE) + DRAM_SIZE_REG_TO_MB(GET_REG(miu1, dram_size)));
#else
    SET_REG(gpu, miu1_start, 0xffff);
    SET_REG(gpu, miu1_end, 0xffff);
#endif

    /* enable RIU */
#ifdef MSTAR_RIU_ENABLED
    SET_REG(gpu, riu, 1);
#endif

    /* set read request length to 64 */
    SET_REG(gpu, rreq_rd_len, 0x2);

    /* set mfdec read request length to 64 */
    SET_REG(gpu, mfdec_rreq_rd_len, 0x2);

    /* enable dynamic clock gating for gpu bridge */
    SET_REG(gpu, dynamic_clk_gating_en, 1);

    /* power on G3DPLL */
    SET_REG(gpu_pll, g3d_pll, 0);
    udelay(1);

    /* return GPU SRAM access right back to GPU */
    CLEAR_REG(chip, bootsram_sel_gpul2, 0);
}

void power_on(void)
{
    SET_REG(clkgen1, ckg_gpu, 0);   /* enable GPU clock */
    SET_REG(gpu, sram_sd_en, 0);    /* power on SRAM */

    /* reset: 0 -> 1 -> 0 */
    SET_REG(gpu, soft_reset, 0);
    SET_REG(gpu, soft_reset, 1);
    udelay(1); /* delay for runtime suspend */
    SET_REG(gpu, soft_reset, 0);
    udelay(1);
}

void power_off(void)
{
    SET_REG(gpu, soft_reset, 1);    /* reset GPU */
    SET_REG(gpu, sram_sd_en, 1);    /* power off SRAM */
    SET_REG(clkgen1, ckg_gpu, 1);   /* disable GPU clock */
    udelay(1);
}

void soft_reset(void)
{
    /* reset GPU hardcore & external GPU bridge circuit */
    SET_REG(gpu, soft_reset, 1);
    udelay(1);
    SET_REG(gpu, soft_reset, 0);

    /* GPU secure sw reset will clear one-way register (0x5b[0], reg_gpu_dram_mask_en)
     * and let GPU use normal AID, which can only R/W normal MIU range.
     */
    SET_REG(gpu, secure_sw_reset, 1);
    udelay(1);
    SET_REG(gpu, secure_sw_reset, 0);
}

#ifdef MSTAR_DISABLE_SHADER_CORES
int get_num_disabled_cores(void)
{
    return 0;
}
#endif

#if defined(MSTAR_M7621) && defined(MSTAR_DEGLITCH_PATCH)
int set_deglitch_mux_to_216mhz(u8 enable)
{
    SET_REG(clkgen1, deglitch_mux, enable > 0 ? 1 : 0 );   /* 1: set deglitch mux to 216Mhz; 0: set deglitch mux to PLL clock */
    return 1;
}
#endif

void get_dram_length(u32* miu0_length, u32* miu1_length, u32* miu2_length)
{
#define DRAM_SIZE_REG_TO_MB(reg)    ((reg > 0) ? (1 << (reg)) : 0)      /* in MB */

    *miu0_length = DRAM_SIZE_REG_TO_MB(GET_REG(miu0, dram_size));
#if defined(MSTAR_MIU1_BUS_BASE) && defined(ARM_MIU1_BASE_ADDR)
    *miu1_length = DRAM_SIZE_REG_TO_MB(GET_REG(miu1, dram_size));
#else
    *miu1_length = 0;
#endif
    *miu2_length = 0;

    return;
}
